from .base import MetricsLayerBase
from .join import Join


class Explore(MetricsLayerBase):
    def __init__(self, definition: dict = {}, project=None) -> None:
        if definition.get("from") is not None:
            definition["from_"] = definition["from"]
        elif definition.get("view_name") is not None:
            definition["from_"] = definition["view_name"]
        else:
            definition["from_"] = definition["name"]

        self.project = project
        self.validate(definition)
        super().__init__(definition)

    @property
    def week_start_day(self):
        if self.model.week_start_day:
            return self.model.week_start_day.lower()
        return "monday"

    def validate(self, definition: dict):
        required_keys = ["name", "model", "from_"]
        for k in required_keys:
            if k not in definition:
                raise ValueError(f"Explore missing required key {k}")

    def validate_fields(self):
        errors = []
        for view_name in self.view_names():
            view = self.project.get_view(view_name)
            referenced_fields = view.referenced_fields()

            errors.extend(
                [
                    f"Could not locate reference {field} in view {view_name} in explore {self.name}"
                    for field in referenced_fields
                    if isinstance(field, str)
                ]
            )
        return list(sorted(list(set(errors))))

    def view_names(self):
        return [self.from_] + [j.name for j in self.joins()]

    def get_join(self, join_name: str):
        return next((j for j in self.joins() if j.name == join_name), None)

    def joins(self):
        output = []
        for j in self._definition.get("joins", []):
            join = Join(j, explore=self, project=self.project)
            if join.is_valid():
                output.append(join)
        return output



from metrics_layer.core.utils import lazy_import

lazy_click = lazy_import("click")

from .seeding import SeedMetricsLayer


@lazy_click.group()
def cli_group():
    pass


@cli_group.command()
def init(arg=None, opt={}):
    """Initialize a metrics layer project"""
    SeedMetricsLayer._init_directories()


@cli_group.command()
@lazy_click.option("--connection", default=None, help="The name of the connection to use for the database")
@lazy_click.option("--database", help="The name of the database to use for seeding")
@lazy_click.option("--schema", default=None, help="The name of the schema to use for seeding")
@lazy_click.argument("profile")
def seed(profile, connection, database, schema):
    """Seed a metrics layer project by referencing the existing database"""
    SeedMetricsLayer._init_directories()
    seeder = SeedMetricsLayer(profile, connection, database, schema)
    seeder.seed()


@cli_group.command()
@lazy_click.argument("profile")
def validate(profile):
    """Validate a metrics layer project, internally, without hitting the database"""
    metrics_layer = SeedMetricsLayer._init_profile(profile)
    errors = metrics_layer.config.project.validate()

    if len(errors) == 0:
        n_explores = len(metrics_layer.config.project.explores())
        lazy_click.echo(f"Project passed (checked {n_explores} explore{'s' if n_explores > 1 else ''})!")
    else:
        lazy_click.echo(f"Found {len(errors)} error{'s' if len(errors)> 1 else ''} in the project:\n")
        for error in errors:
            lazy_click.echo(f"\n{error}\n")


if __name__ == "__main__":
    cli_group()



from metrics_layer.cli import cli_group  # noqa
from metrics_layer.core import MetricsLayerConnection  # noqa
from metrics_layer.core.utils import lazy_import

try:
    import importlib.metadata as lazy_importlib_metadata
except ModuleNotFoundError:
    lazy_importlib_metadata = lazy_import("importlib_metadata")

__version__ = lazy_importlib_metadata.version(__name__)



import json
import os


class SeedMetricsLayer:
    def __init__(self, profile, connection=None, database=None, schema=None):
        self.profile_name = profile
        self.metrics_layer, self.connection = self._init_connection(self.profile_name, connection)
        self.database = database
        self.schema = schema

        self._snowflake_type_lookup = {
            "DATE": "date",
            "DATETIME": "timestamp",
            "TIMESTAMP_TZ": "timestamp",
            "TIMESTAMP_NTZ": "timestamp",
            "TIMESTAMP_LTZ": "timestamp",
            "TIMESTAMP": "timestamp",
            "BOOLEAN": "yesno",
            "FIXED": "number",
            "FLOAT": "number",
            "REAL": "number",
            "TEXT": "string",
            "VARCHAR": "string",
            "CHAR": "string",
            "CHARACTER": "string",
            "STRING": "string",
            "TEXT": "string",
            "BINARY": "string",
            "VARBINARY": "string",
        }

    def seed(self):
        import pandas as pd

        from metrics_layer.core.model.definitions import Definitions
        from metrics_layer.core.parse.project_reader import ProjectReader

        if self.connection.type != Definitions.snowflake:
            raise NotImplementedError(
                "The only data warehouse supported for seeding at this time is Snowflake"
            )
        table_query = self.table_query(type="tables")
        table_data = self.run_query(table_query)
        view_query = self.table_query(type="views")
        view_data = self.run_query(view_query)

        columns = ["DATABASE_NAME", "SCHEMA_NAME", "NAME"]
        data = pd.concat([table_data[columns], view_data[columns]], sort=False)

        # Each row represents either a single table or a single view
        views = []
        for _, row in data.iterrows():
            columns_query = self.columns_query(row["NAME"], row["SCHEMA_NAME"], row["DATABASE_NAME"])
            column_data = self.run_query(columns_query)

            column_data["RAW_TYPE"] = column_data["DATA_TYPE"].apply(
                lambda x: json.loads(x).get("type", "TEXT")
            )

            view = self.make_view(column_data, row["NAME"], row["SCHEMA_NAME"])
            print(view)
            views.append(view)

        models = self.make_models(views)

        reader = ProjectReader(None)
        # Fake that the project was loaded from a repo
        reader.unloaded = False

        print(models)
        reader._models = models
        reader._views = views
        # Dump the models to yaml files
        reader.dump(os.getcwd())

    def make_models(self, views: list):
        model = {
            "version": 1,
            "type": "model",
            "name": "base_model",
            "connection": self.connection.name,
            "explores": [{"name": view["name"]} for view in views],
        }
        return [model]

    def make_view(self, column_data, table_name: str, schema_name: str):
        view_name = self.clean_name(table_name)
        count_measure = {"field_type": "measure", "name": "count", "type": "count"}
        fields = self.make_fields(column_data) + [count_measure]
        view = {
            "version": 1,
            "type": "view",
            "name": view_name,
            "sql_table_name": f"{schema_name}.{table_name}",
            "default_date": next((f["name"] for f in fields if f["field_type"] == "dimension_group"), None),
            "row_label": "TODO - Label row",
            "fields": fields,
        }
        return view

    def make_fields(self, column_data: str):
        fields = []
        for _, row in column_data[["COLUMN_NAME", "RAW_TYPE"]].iterrows():
            name = self.clean_name(row["COLUMN_NAME"])
            metrics_layer_type = self._snowflake_type_lookup.get(row["RAW_TYPE"], "string")
            sql = "${TABLE}." + row["COLUMN_NAME"]

            field = {"name": name, "sql": sql}

            if metrics_layer_type in {"timestamp", "date"}:
                field["field_type"] = "dimension_group"
                field["type"] = "time"
                field["timeframes"] = ["raw", "date", "week", "month", "quarter", "year"]
                field["datatype"] = metrics_layer_type
            else:
                field["field_type"] = "dimension"
                field["type"] = metrics_layer_type
            fields.append(field)
        return fields

    def table_query(self, type="tables"):
        query = f"show {type} in "
        if self.database and not self.schema:
            query += f"database {self.database};"
        elif self.database and self.schema:
            query += f"schema {self.database}.{self.schema};"
        else:
            raise ValueError("You must specify a database or a database and a schema for seeding")
        return query

    def columns_query(self, table_name: str, schema_name: str, database_name: str):
        return f'show columns in "{database_name}"."{schema_name}"."{table_name}";'

    def run_query(self, query: str):
        return self.metrics_layer.run_query(query, self.connection)

    @staticmethod
    def _init_connection(profile_name: str, connection_name: str = None):
        metrics_layer = SeedMetricsLayer._init_profile(profile_name)

        if connection_name:
            connection = metrics_layer.config.get_connection(connection_name)
        else:
            connections = metrics_layer.config.connections()
            if len(connections) == 1:
                connection = connections[0]
            else:
                raise ValueError(
                    f"Could not determine the connection to use with profile {profile_name}, "
                    "please pass the connection name with the --connection arg"
                )
        return metrics_layer, connection

    @staticmethod
    def _init_profile(profile_name: str):
        from metrics_layer.core import MetricsLayerConnection

        metrics_layer = MetricsLayerConnection(profile_name)
        return metrics_layer

    @staticmethod
    def _init_directories():
        models_dir = "models/"
        views_dir = "views/"
        for directory in [models_dir, views_dir]:
            fully_qualified_path = os.path.join(os.getcwd(), directory)
            if not os.path.exists(fully_qualified_path):
                os.mkdir(fully_qualified_path)

    @staticmethod
    def clean_name(txt: str):
        return txt.lower().replace(" ", "_")



import re


class MetricsLayerBase:
    def __init__(self, definition: dict = {}) -> None:
        self._definition = definition

    def __getattr__(self, attr: str):
        return self._definition.get(attr, None)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"

    def to_dict(self):
        return self._definition


class SQLReplacement:
    @staticmethod
    def fields_to_replace(text: str):
        matches = re.finditer(r"\$\{(.*?)\}", text, re.MULTILINE)
        return [match.group(1) for match in matches]
